/*
 *  hnv_main_X11.c
 *  HeteroNumberViewer
 *
 *  Created by Naoto Morikawa on 9/9/07.
 *  Copyright (C) 2007 Naoto Morikawa. All rights reserved.
 *
 *  2007-10-01 version 0.1 release
 */

#include <stdlib.h>
// #include <GLUT/glut.h>
#include <GL/glut.h>
#include <stdio.h>
#include <math.h>
#include "hnv_main.h"
#include "hnv_computation_main.h"
#include "hnv_set_up_tetra_flow.h"
#include "hnv_my_init.h"
#include "hnv_scene_flow.h"
#include "hnv_tetrahedron.h"
#include "hnv_closure.h"

//===============================================
// Private data
//===============================================

static int hnv_rot = OFF;
static GLfloat Xrot = 0.0, Yrot = 0.0;
static int hnv_trans = OFF;
static GLfloat CamX = 0.0, CamY = 0.0;
static float ScaleFactor = 1.0;
static int hnv_blend = ON;
static float hnv_zoom_in = 1;
static int hnv_closure = OFF;
static int hnv_visibility = ON;
static int hnv_visible_index = -1;
static int hnv_peak_visibility = ON;
static int hnv_visible_peak_index = -1;
static HNV_VISIBLE_PEAKS hnv_visible_peaks = HNV_STD_CONE;
static HNV_TRAJ_TYPE hnv_visible_traj_type = HNV_CLOSED;
const char * work_filename;		// used in keyboard()

//===============================================
// Public data
//===============================================

//===============================================
// Private methods
//===============================================
//
// void display(void);
// void reshape(int width, int height);
// void keyboard(unsigned char key, int x, int y);
// static void MouseMove(int x, int y);
// void idle(void);
//-----------------------------------------------

void display(void)
{
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
	
	glColor3f (1.0, 1.0, 1.0);
	glLoadIdentity ();
	glTranslatef (CamX, CamY, -15.0);	// Camera position
	
	ScaleFactor = 12.0 * hnv_zoom_in / Ssize[3];
	glScalef(ScaleFactor, ScaleFactor, ScaleFactor);
	
	glPushMatrix();
	glRotatef(Xrot, 1, 0, 0);
	glRotatef(Yrot, 0, 1, 0);
	// Tcenter[] = The center of the peaks
	glTranslatef(Tcenter[0], Tcenter[1], Tcenter[2]);
	
	if (hnv_closure == OFF) {
		scene_flow(pt_my_tetrahedron_flow, 
			   hnv_visibility, hnv_visible_index, 
			   hnv_peak_visibility, hnv_visible_peak_index, 
			   hnv_visible_peaks, hnv_visible_traj_type);
	} else {
		scene_flow(&my_closure_tetrahedron_flow, 
			   hnv_visibility, hnv_visible_index, 
			   hnv_peak_visibility, hnv_visible_peak_index, 
			   hnv_visible_peaks, hnv_visible_traj_type);
	}
	
	glPopMatrix();
	
	glFlush();
	glutSwapBuffers();
}


void reshape(int width, int height)
{
	glViewport (0, 0, (GLsizei) width, (GLsizei) height); 
	glMatrixMode (GL_PROJECTION);
	glLoadIdentity ();
	glFrustum (-1.0, 1.0, -1.0, 1.0, 1.5, 20.0);
	glMatrixMode (GL_MODELVIEW);
	glLoadIdentity ();
	glTranslatef(0.0,0.0,-1.0);
}


void keyboard(unsigned char key, int x, int y)
{
	switch (key) {
		case 'B':
			if (hnv_blend < 0) {	// OFF
				hnv_blend = ON;
				glEnable(GL_BLEND);
				glDisable(GL_DEPTH_TEST);
			} else {
				hnv_blend = OFF;
				glDisable(GL_BLEND);
				glEnable(GL_DEPTH_TEST);
			}
			break;
		case 'C':	// Closure of the tetrahedron flow
			if (hnv_closure == ON) {
				fprintf(STDOUT, "HNV> ERROR (The current flow is the closure)\n");
				break;
			}
			if (check_tflow_seq_array(&my_closure_tetrahedron_flow) == FAIL) {
				if (get_tetra_tetra_ct(&bd_tetra_list_head) > 0) {
					int i;
					
					if (0) {
						fprintf(STDOUT, "keyboard() : DEBUG START of bd_tetra_list_head\n");
						print_tetra_list(&bd_tetra_list_head);
						fprintf(STDOUT, "keyboard() : DEBUG END of bd_tetra_list_head\n");
						fprintf(STDOUT, "\n");
					}
					hnv_closure = OFF;
					compute_closure();
					i = 1;
					while (get_tetra_tetra_ct(&bd_tetra_list_head) > 0) {
						if (0)
							fprintf(STDERR, "keyboard() : DEBUG result of %d-th closure => %d BD tetrahedrons\n", 
								i, get_tetra_tetra_ct(&bd_tetra_list_head));
						if (1) {
							fprintf(STDOUT, "HNV> Summary of the %d-th closure\n", i);
							fprintf(STDOUT, "\n");
							print_out_result_summary(STDOUT, &my_closure_tetrahedron_flow);
							fprintf(STDOUT, "\n");
							fprintf(STDOUT, "HNV>\n");
						}
						compute_closure();
						i++;
					}
					hnv_closure = ON;
					fprintf(STDOUT, "HNV> Show the closure of the current tetrahedron flow\n");
				} else {
					fprintf(STDERR, "HNV> The current flow is already closed (no boundary tetrahedrons)\n");
				}
			} else {	// already computed
				hnv_closure = ON;
				fprintf(STDOUT, "HNV> Show the closure of the current tetrahedron flow (saved data)\n");
			}
			break;
		case 'c':	// the original flow
			if (hnv_closure == OFF) {
				fprintf(STDOUT, "HNV> ERROR (The current tetrahedron flow is the original flow)\n");
				break;
			}
			hnv_closure = OFF;
			fprintf(STDOUT, "HNV> Show the current tetrahedron flow\n");
			break;
		case 'H':	// help
			print_help();
			break;
		case 'L':	// Reload (default) the file        <=> 'l'
			if (hnv_closure == ON) {
				fprintf(STDOUT, "HNV> ERROR (Closure data couldn't be overwritten. Press 'c' then 'L')\n");
				break;
			}
			set_use_extended_conj_cone(HNV_FALSE);
			set_remove_redundant_peaks(HNV_FALSE);
			pt_my_tetrahedron_flow = pt_my_tetrahedron_flow_org;
			if (check_tflow_seq_array(pt_my_tetrahedron_flow_new) == SUCCESS) {
				free_tetra_flow(pt_my_tetrahedron_flow_new);
			}
			if (check_tflow_seq_array(&my_closure_tetrahedron_flow) == SUCCESS) {
				hnv_closure = OFF;
				free_tetra_flow(&my_closure_tetrahedron_flow);
			}
			free_cones_and_tetras();
			
			compute_tetrahedron_flow(work_filename);
			set_up_tetra_flow(pt_my_tetrahedron_flow_new);
			pt_my_tetrahedron_flow = pt_my_tetrahedron_flow_new;
			// set_up_tcenter_and_ssize(pt_my_tetrahedron_flow);
			fprintf(STDOUT, "HNV> Reload the input file (default)\n");
			break;
		case 'l':	// Show the saved tetrahedron flow  <=> 'L' or 'X'
			if (hnv_closure == ON) {
				fprintf(STDOUT, "HNV> ERROR (Shown is the closure data. Press 'c' then 'l')\n");
				break;
			}
			pt_my_tetrahedron_flow = pt_my_tetrahedron_flow_org;
			fprintf(STDOUT, "HNV> Show the saved tetrahedron flow\n");
			break;
		case 'O':	// Print out the .tetra file of the current flow
			if (hnv_closure == ON) {
				print_out_result(work_filename, &my_closure_tetrahedron_flow);
			} else {
				print_out_result(work_filename, pt_my_tetrahedron_flow);
			}
			break;
		case 'o':	// Print out the current tetrahedron flow
			if (hnv_closure == ON) {
				fprintf(STDOUT, "HNV> Summary of the closure\n");
				fprintf(STDOUT, "\n");
				print_out_result_summary(STDOUT, &my_closure_tetrahedron_flow);
				fprintf(STDOUT, "\n");
				fprintf(STDOUT, "HNV>\n");
			} else {
				fprintf(STDOUT, "HNV> Summary of the current tetrahedron flow\n");
				fprintf(STDOUT, "\n");
				print_out_result_summary(STDOUT, pt_my_tetrahedron_flow);
				fprintf(STDOUT, "\n");
				fprintf(STDOUT, "HNV>\n");
			}
			break;
		case 'P':	// to show peaks of std_cone/conj_roof/std_roof (toggle)
			if (hnv_visible_peaks == HNV_STD_CONE) {
				if (hnv_peak_visibility == ON) {
					hnv_visible_peaks = HNV_CONJ_ROOF;
					fprintf(STDOUT, "HNV> Show the peaks of the (extended) conjugate roof\n");
				} else {
					int peak_ct;
					
					hnv_peak_visibility = ON;
					fprintf(STDOUT, "HNV> Show the peaks of the standard cone\n");
					
					if (hnv_closure == ON) {
						peak_ct = get_tflow_std_cone_peak_size(&my_closure_tetrahedron_flow);
					} else {
						peak_ct = get_tflow_std_cone_peak_size(pt_my_tetrahedron_flow);
					}
					if (peak_ct == 0) break;
					hnv_visible_peak_index = (hnv_visible_peak_index + peak_ct - 1) % peak_ct;
				}
			} else if (hnv_visible_peaks == HNV_CONJ_ROOF) {
				hnv_visible_peaks = HNV_STD_ROOF;
				fprintf(STDOUT, "HNV> Show the peaks of the standard roof\n");
			} else {
				hnv_visible_peaks = HNV_STD_CONE;
				fprintf(STDOUT, "HNV> Show the peaks of the standard cone\n");
			}
			break;
		case 'p':	// shift the visible peak of the standard cone
			if (hnv_visible_peaks == HNV_STD_CONE) {
				int peak_ct;
				
				if (hnv_closure == ON) {
					peak_ct = get_tflow_std_cone_peak_size(&my_closure_tetrahedron_flow);
				} else {
					peak_ct = get_tflow_std_cone_peak_size(pt_my_tetrahedron_flow);
				}
				if (peak_ct == 0) break;
				hnv_peak_visibility = OFF;
				hnv_visible_peak_index = (hnv_visible_peak_index + 1) % peak_ct;
				fprintf(STDOUT, "HNV> Show the %d-th of %d peaks of the standard cone\n",
					hnv_visible_peak_index, peak_ct);
			} else {
				fprintf(STDOUT, "HNV> ERROR (Not the peaks of the standard cone)\n");
			}
			break;
		case 'Q':
			fprintf(STDOUT, "HNV> quit\n");
			exit(0);
		case 'R':	// Rotation
			if (hnv_rot > OFF) {
				hnv_rot = OFF;
			} else {
				hnv_rot = ON;
				hnv_trans = OFF;
			}
			break;
		case 'S':	// Save the current tetrahedron_flow
			if (hnv_closure == ON) {
				fprintf(STDOUT, "HNV> ERROR (Closure data couldn't be saved. Press 'c' then 'S')\n");
				break;
			}
			if (pt_my_tetrahedron_flow == pt_my_tetrahedron_flow_new) {
				struct tetra_flow * tmp_t_flow;
				
				tmp_t_flow = pt_my_tetrahedron_flow_org;
				pt_my_tetrahedron_flow_org = pt_my_tetrahedron_flow_new;
				pt_my_tetrahedron_flow_new = tmp_t_flow;
				pt_my_tetrahedron_flow = pt_my_tetrahedron_flow_org;
				
				if (check_tflow_seq_array(&my_closure_tetrahedron_flow) == SUCCESS) {
					hnv_closure = OFF;
					free_tetra_flow(&my_closure_tetrahedron_flow);
				}	
			}
			fprintf(STDOUT, "HNV> Save the current tetrahedron flow\n");
			break;
		case 'T':	// Translation
			if (hnv_trans > OFF) {
				hnv_trans = OFF;
			} else {
				hnv_trans = ON;
				hnv_rot = OFF;
			}
			break;
		case 'U':	// Undo all
			Xrot = 0.0;
			Yrot = 0.0;
			CamX = 0.0;
			CamY = 0.0;
			ScaleFactor = 1.0;
			Tcenter[0] = Tcenter_save[0];
			Tcenter[1] = Tcenter_save[1];
			Tcenter[2] = Tcenter_save[2];
			hnv_blend = ON;
			hnv_zoom_in = 1;
			hnv_rot = OFF;
			hnv_trans = OFF;
			hnv_closure = OFF;
			hnv_visibility = ON;
			hnv_visible_index = 0;
			hnv_visible_peaks = HNV_STD_CONE;
			hnv_visible_traj_type = HNV_CLOSED;
			fprintf(STDOUT, "HNV> Undo all\n");
			fprintf(STDOUT, "HNV> Show open trajectories and the peaks of the standard cone\n");
			break;
		case 'V':	// show one/all trajectories
			if (hnv_visibility == ON) {
				hnv_visibility = OFF;
				fprintf(STDOUT, "HNV> Show single trajectory\n");
			} else {
				hnv_visibility = ON;
				fprintf(STDOUT, "HNV> Show multiple trajectories\n");
			}
			break;
		case 'v':	// shift the visible trajectory
		{
			int seq_ct;
			
			if (hnv_closure == ON) {
				seq_ct = get_tflow_seq_ct(&my_closure_tetrahedron_flow);
			} else {
				seq_ct = get_tflow_seq_ct(pt_my_tetrahedron_flow);
			}
			if (seq_ct == 0) break;
			hnv_visibility = OFF;
			hnv_visible_index = (hnv_visible_index + 1) % seq_ct;
			fprintf(STDOUT, "HNV> Show the %d-th of %d trajectories\n",
				hnv_visible_index, seq_ct);
		}
			break;
		case 'W':	// show closed/open/no trajectories (toggle)
			hnv_visibility = ON;
			if (hnv_visible_traj_type == HNV_CLOSED) {
				hnv_visible_traj_type = HNV_NOT_CLOSED;
				fprintf(STDOUT, "HNV> Show open trajectories\n");
			} else if (hnv_visible_traj_type == HNV_NOT_CLOSED) {
				hnv_visible_traj_type = HNV_UNKNOWN_T;
				fprintf(STDOUT, "HNV> Show no trajectories\n");
			} else if (hnv_visible_traj_type == HNV_UNKNOWN_T) {
				hnv_visible_traj_type = HNV_ALL;
				fprintf(STDOUT, "HNV> Show all trajectories\n");
			} else {
				hnv_visible_traj_type = HNV_CLOSED;
				fprintf(STDOUT, "HNV> Show closed trajectories\n");
			}
			break;
		case 'X':	// Reload (the 'x' option) the file        <=> 'l'
			if (hnv_closure == ON) {
				fprintf(STDOUT, "HNV> ERROR (Closure data couldn't be overwritten. Press 'c' then 'X')\n");
				break;
			}
			set_use_extended_conj_cone(HNV_TRUE);
			set_remove_redundant_peaks(HNV_TRUE);
			pt_my_tetrahedron_flow = pt_my_tetrahedron_flow_org;
			if (check_tflow_seq_array(pt_my_tetrahedron_flow_new) == SUCCESS) {
				free_tetra_flow(pt_my_tetrahedron_flow_new);
			}
			if (check_tflow_seq_array(&my_closure_tetrahedron_flow) == SUCCESS) {
				hnv_closure = OFF;
				free_tetra_flow(&my_closure_tetrahedron_flow);
			}
			free_cones_and_tetras();
			
			compute_tetrahedron_flow(work_filename);
			set_up_tetra_flow(pt_my_tetrahedron_flow_new);
			pt_my_tetrahedron_flow = pt_my_tetrahedron_flow_new;
			// set_up_tcenter_and_ssize(pt_my_tetrahedron_flow);
			fprintf(STDOUT, "HNV> Reload the input file (the 'x' option)\n");
			break;
		case 'Z':	// Zoom in
			hnv_zoom_in = hnv_zoom_in * 5/4;
			break;
		case 'z':	// Zoom out
			hnv_zoom_in = hnv_zoom_in * 4/5;
			break;
		default:
			break;
	}
	return;
}

/*  -----------------
*   y-axis
*   ^
*   |
*  (0,0)  ->  x-axis
*  -----------------
* (x,y,0): ==> (tmp_a,tmp_b,tmp_c) 
* glRotatef(-Xrot,1,0,0): ==> (x, cos(-X)y, sin(-X)y)
* glRotatef(-Yrot,0,1,0): ==> (cos(-Y)x - sin(-Y)sin(-X)y, cos(-X)y,
			       *				sin(-Y)x - cos(-Y)sin(-X)y)
*/
static void MouseMove(int x, int y)
{
	if (hnv_rot == ON) {
		Yrot = x - WindowWidth/2;
		Xrot = y - WindowHeight/2;
	} else if (hnv_trans == ON) {
		CamX = (x - WindowWidth/2) / (100.0*ScaleFactor);
		CamY = -(y - WindowHeight/2)  / (100.0*ScaleFactor);
	}
	glutPostRedisplay();
	return;
}


void idle(void)
{
	glutPostRedisplay();
}


//===============================================
// Public methods
//===============================================

int main(int argc, char** argv)
{	
	/* The last arg is the filename */
	work_filename = argv[argc -1];		// used in keyboard()
	
	glutInit(&argc, argv);
	glutInitDisplayMode(GLUT_RGBA | GLUT_DOUBLE | GLUT_DEPTH);
	glutInitWindowSize(WindowWidth, WindowHeight);
	glutCreateWindow("Hetero Number Viewer");
    
	my_init(argc, argv);

	glutDisplayFunc(display);
	glutReshapeFunc(reshape);
	glutIdleFunc(idle);
	glutKeyboardFunc(keyboard);
	glutPassiveMotionFunc(MouseMove);
	glutMainLoop();

	my_fin();
    
	return EXIT_SUCCESS;
}

