/*
 *  hnv_scene_flow_X11.c
 *  HeteroNumberViewer
 *
 *  Created by Naoto Morikawa on 9/17/07.
 *  Copyright (C) 2007 Naoto Morikawa. All rights reserved.
 *
 *  2007-10-01 version 0.1 release
 *  2007-10-14 hnv_scene_flow.c: draw BD tetrahderons in closed trajectories as 
 *                               yellow tetrahedron in scene_flow()
 *  2007-10-14 version 0.3 release
 */

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include "hnv_main.h"
#include "hnv_set_up_tetra_flow.h"
#include "hnv_t_flow_data_strct_X11.h"	// contains #include <GLUT/glut.h>

//===============================================
// Private data
//===============================================

GLfloat white[] = {1.0, 1.0, 1.0, 0.30};
GLfloat aqua_blue[] = {0.8, 0.7, 0.8, 0.75};
GLfloat light_blue[] = {0.6, 0.7, 1.0, 0.75};
GLfloat blue[] = {0.1, 0.5, 1.0, 0.75};
GLfloat red[] = {1.0, 0.2, 0.2, 0.75};
GLfloat yellow[] = {1.0, 0.9, 0.3, 0.75};

GLfloat bright_shininess[] = {120.0};	// [0.0, 128.0]
GLfloat dark_shininess[] = {80.0};

GLfloat * tetra_closed_color = light_blue;
GLfloat * closed_shininess = bright_shininess;
GLfloat * tetra_open_color = aqua_blue;
GLfloat * open_shininess = bright_shininess;
GLfloat * tetra_bd_color = yellow;
GLfloat * bd_shininess = bright_shininess;
GLfloat * stick_closed_color = white;
GLfloat * stick_open_color = white;
GLfloat * stick_shininess = bright_shininess;
GLfloat * sphere_shininess = bright_shininess;


//===============================================
// Public data
//===============================================

//===============================================
// Private methods
//===============================================
// 
// void draw_tetra_tile_left(struct tetra_tile * cur_tetra, GLfloat * color,
//			     GLfloat * shininess);
// void draw_tetra_tile_right(struct tetra_tile * cur_tetra, GLfloat * color,
//			      GLfloat * shininess);
// void draw_cylinder_stick(struct tetra_tile * cur_tile, GLfloat radius);
// void draw_peak_sphere(struct tetra_peak * cur_peak, GLfloat * color,
//			 GLfloat * shininess, GLfloat radius)
//-----------------------------------------------

void draw_tetra_tile_right(struct tetra_tile * cur_tetra, GLfloat * diffuse,
		     GLfloat * specular, GLfloat * shininess)
{
	glPushAttrib(GL_CURRENT_BIT);
	glMaterialfv(GL_FRONT, GL_DIFFUSE, diffuse);
	glMaterialfv(GL_FRONT, GL_SPECULAR, specular);
	glMaterialfv(GL_FRONT, GL_SHININESS, shininess);
	glBegin(GL_TRIANGLE_STRIP);
	glNormal3f(cur_tetra->normals[0].coor[0],
		   cur_tetra->normals[0].coor[1],
		   cur_tetra->normals[0].coor[2]);
	glVertex3f(cur_tetra->vertices[0].coor[0],
		   cur_tetra->vertices[0].coor[1],
		   cur_tetra->vertices[0].coor[2]);
	glNormal3f(cur_tetra->normals[1].coor[0],
		   cur_tetra->normals[1].coor[1],
		   cur_tetra->normals[1].coor[2]);
	glVertex3f(cur_tetra->vertices[1].coor[0],
		   cur_tetra->vertices[1].coor[1],
		   cur_tetra->vertices[1].coor[2]);
	glNormal3f(cur_tetra->normals[2].coor[0],
		   cur_tetra->normals[2].coor[1],
		   cur_tetra->normals[2].coor[2]);
	glVertex3f(cur_tetra->vertices[2].coor[0],
		   cur_tetra->vertices[2].coor[1],
		   cur_tetra->vertices[2].coor[2]);
	glNormal3f(cur_tetra->normals[3].coor[0],
		   cur_tetra->normals[3].coor[1],
		   cur_tetra->normals[3].coor[2]);
	glVertex3f(cur_tetra->vertices[3].coor[0],
		   cur_tetra->vertices[3].coor[1],
		   cur_tetra->vertices[3].coor[2]);
	glNormal3f(cur_tetra->normals[0].coor[0],
		   cur_tetra->normals[0].coor[1],
		   cur_tetra->normals[0].coor[2]);
	glVertex3f(cur_tetra->vertices[0].coor[0],
		   cur_tetra->vertices[0].coor[1],
		   cur_tetra->vertices[0].coor[2]);
	glNormal3f(cur_tetra->normals[1].coor[0],
		   cur_tetra->normals[1].coor[1],
		   cur_tetra->normals[1].coor[2]);
	glVertex3f(cur_tetra->vertices[1].coor[0],
		   cur_tetra->vertices[1].coor[1],
		   cur_tetra->vertices[1].coor[2]);
	glEnd();
	glPopAttrib();
	return;
}


void draw_tetra_tile_left(struct tetra_tile * cur_tetra, GLfloat * diffuse,
			  GLfloat * specular, GLfloat * shininess)
{
	glPushAttrib(GL_CURRENT_BIT);
	glMaterialfv(GL_FRONT, GL_DIFFUSE, diffuse);
	glMaterialfv(GL_FRONT, GL_SPECULAR, specular);
	glMaterialfv(GL_FRONT, GL_SHININESS, shininess);
	glBegin(GL_TRIANGLE_STRIP);
	glNormal3f(cur_tetra->normals[0].coor[0],
		   cur_tetra->normals[0].coor[1],
		   cur_tetra->normals[0].coor[2]);
	glVertex3f(cur_tetra->vertices[0].coor[0],
		   cur_tetra->vertices[0].coor[1],
		   cur_tetra->vertices[0].coor[2]);
	glNormal3f(cur_tetra->normals[2].coor[0],
		   cur_tetra->normals[2].coor[1],
		   cur_tetra->normals[2].coor[2]);
	glVertex3f(cur_tetra->vertices[2].coor[0],
		   cur_tetra->vertices[2].coor[1],
		   cur_tetra->vertices[2].coor[2]);
	glNormal3f(cur_tetra->normals[1].coor[0],
		   cur_tetra->normals[1].coor[1],
		   cur_tetra->normals[1].coor[2]);
	glVertex3f(cur_tetra->vertices[1].coor[0],
		   cur_tetra->vertices[1].coor[1],
		   cur_tetra->vertices[1].coor[2]);
	glNormal3f(cur_tetra->normals[3].coor[0],
		   cur_tetra->normals[3].coor[1],
		   cur_tetra->normals[3].coor[2]);
	glVertex3f(cur_tetra->vertices[3].coor[0],
		   cur_tetra->vertices[3].coor[1],
		   cur_tetra->vertices[3].coor[2]);
	glNormal3f(cur_tetra->normals[0].coor[0],
		   cur_tetra->normals[0].coor[1],
		   cur_tetra->normals[0].coor[2]);
	glVertex3f(cur_tetra->vertices[0].coor[0],
		   cur_tetra->vertices[0].coor[1],
		   cur_tetra->vertices[0].coor[2]);
	glNormal3f(cur_tetra->normals[2].coor[0],
		   cur_tetra->normals[2].coor[1],
		   cur_tetra->normals[2].coor[2]);
	glVertex3f(cur_tetra->vertices[2].coor[0],
		   cur_tetra->vertices[2].coor[1],
		   cur_tetra->vertices[2].coor[2]);
	glEnd();
	glPopAttrib();
	return;
}


void draw_cylinder_stick(struct tetra_tile * cur_tile, GLfloat radius,
			 GLfloat * diffuse, GLfloat * specular, GLfloat * shininess)
{
	float dTheta, Theta;
	float axis_x, axis_y, axis_z;	// axis of rotation
	float angle;			// angle of rotation
	int i;
	
	dTheta = 2 * PI / 20.0;
	Theta = dTheta;
	
	// direction of the initial cylinder =  { 0.0, 1.0, 0.0}
	// axis of rotation = direction of "{ 0.0, 1.0, 0.0} x {dir_vex->x, dir_vex->y, dir_vex->z}"
	// where A x B = (Ay*Bz - Az*By, Az*Bx - Ax*Bz, Ax*By - Ay*Bx)
	axis_x = cur_tile->direction_vector.coor[2];
	axis_y = 0.0;
	axis_z = - cur_tile->direction_vector.coor[0];
	// (p,q) = |p||q|cos(angle)
	angle = (float) (180.0/PI) * acos(cur_tile->direction_vector.coor[1]);
	
	glPushMatrix();
	glPushAttrib(GL_CURRENT_BIT);
		glTranslatef(cur_tile->vertices[0].coor[0],
			     cur_tile->vertices[0].coor[1],
			     cur_tile->vertices[0].coor[2]);
		if ((axis_x == 0.0) && (axis_z == 0.0)){
			fprintf(STDERR, 
				"draw_cylinder_stick() : ERROR axis of rotation == (0, 0, 0)\n");
			exit(1);
		}else {
			glRotatef(angle, axis_x, axis_y, axis_z);
		}

		glMaterialfv(GL_FRONT, GL_DIFFUSE, diffuse);
		glMaterialfv(GL_FRONT, GL_SPECULAR, specular);
		glMaterialfv(GL_FRONT, GL_SHININESS, shininess);
		glutSolidSphere(radius, 10, 10);

		glBegin(GL_QUAD_STRIP);
			glNormal3f(0.0, 0.0, 1.0);
			glVertex3f(0.0F, cur_tile->direction_vector_length, radius);
			glVertex3f(0.0F, 0.0F, radius);
			for (i = 1; i < 20; i++, Theta += dTheta) {
				float cs = (float) cos((double)Theta);
				float sn = (float) sin((double)Theta);
				float x = radius * sn;
				float z = radius * cs;
		
				glNormal3f(sn, 0.0, cs);
				glVertex3f(x, cur_tile->direction_vector_length, z);
				glVertex3f(x, 0.0F, z);
			}
			glNormal3f(0.0, 0.0, 1.0);
			glVertex3f(0.0F, cur_tile->direction_vector_length, radius);
			glVertex3f(0.0F, 0.0F, radius);
		glEnd();
		
	glPopAttrib();
	glPopMatrix();
	
	glPushMatrix();
		glTranslatef(cur_tile->vertices[3].coor[0], 
			     cur_tile->vertices[3].coor[1],
			     cur_tile->vertices[3].coor[2]);
		glutSolidSphere(radius, 10, 10);
	glPopMatrix();
	return;
}


void draw_peak_sphere(struct tetra_peak * cur_peak, GLfloat * color,
		      GLfloat * shininess, GLfloat radius)
{
	glPushMatrix();
	glPushAttrib(GL_CURRENT_BIT);
		glMaterialfv(GL_FRONT, GL_DIFFUSE, color);
		glMaterialfv(GL_FRONT, GL_SPECULAR, color);
		glMaterialfv(GL_FRONT, GL_SHININESS, shininess);
		glTranslatef(cur_peak->coor3d[0], cur_peak->coor3d[1],
			     cur_peak->coor3d[2]);
		glutSolidSphere(radius, 10, 10);
	glPopAttrib();
	glPopMatrix();
	return;
}


//===============================================
// Public methods
//===============================================

#define HNV_RADIUS  0.10
#define HNV_RADIUS_L  0.20
#define HNV_RADIUS_S  0.05
void scene_flow(struct tetra_flow * t_flow, int visibility, int visible_index, 
		int peak_visibility, int visible_peak_index,
		HNV_VISIBLE_PEAKS visible_peak, HNV_TRAJ_TYPE visible_traj_type)
{
	struct tetra_seq * cur_seq;
	struct tetra_tile * cur_tetra;
	int seq_ct;
	int peak_ct;
	int t_ct;
	
	// tetrahedron flow
	if (visibility == ON) {	// Show all trajectories
		for (seq_ct = 0; seq_ct < t_flow->seq_ct; seq_ct++) {
			cur_seq = t_flow->sequences + seq_ct;
			
			for (t_ct = 0; t_ct < cur_seq->length; t_ct++) {
				cur_tetra = cur_seq->tiles + t_ct;
				
				if ( visible_traj_type == HNV_CLOSED) {
					if (cur_seq->type == HNV_CLOSED) {
						if (cur_tetra->belongs_to == HNV_BD) {	// 2007-10-14
							if (cur_tetra->chiral == HNV_RIGHT_CHIRAL) {
								draw_tetra_tile_right(cur_tetra, tetra_bd_color, tetra_bd_color, bd_shininess);
							} else {
								draw_tetra_tile_left(cur_tetra, tetra_bd_color, tetra_bd_color, bd_shininess);
							}
							draw_cylinder_stick(cur_tetra, HNV_RADIUS, tetra_bd_color, tetra_bd_color, stick_shininess);
						} else {
							if (cur_tetra->chiral == HNV_RIGHT_CHIRAL) {
								draw_tetra_tile_right(cur_tetra, tetra_closed_color, tetra_closed_color, closed_shininess);
							} else {
								draw_tetra_tile_left(cur_tetra, tetra_closed_color, tetra_closed_color, closed_shininess);
							}
							draw_cylinder_stick(cur_tetra, HNV_RADIUS, stick_closed_color, stick_closed_color, stick_shininess);
						}
					} else {
						if (cur_tetra->belongs_to == HNV_BD) {
							draw_cylinder_stick(cur_tetra, HNV_RADIUS_S, tetra_bd_color, tetra_bd_color, bd_shininess);
						} else {
							draw_cylinder_stick(cur_tetra, HNV_RADIUS_S, stick_closed_color, stick_closed_color, stick_shininess);
						}
					}
					
				} else if (visible_traj_type == HNV_NOT_CLOSED) {
					if (cur_seq->type == HNV_CLOSED) {
						draw_cylinder_stick(cur_tetra, HNV_RADIUS_S, stick_closed_color, stick_closed_color, stick_shininess);
					} else {
						if (cur_tetra->belongs_to == HNV_BD) {
							if (cur_tetra->chiral == HNV_RIGHT_CHIRAL) {
								draw_tetra_tile_right(cur_tetra, tetra_bd_color, tetra_bd_color, bd_shininess);
							} else {
								draw_tetra_tile_left(cur_tetra, tetra_bd_color, tetra_bd_color, bd_shininess);
							}
							draw_cylinder_stick(cur_tetra, HNV_RADIUS, tetra_bd_color, tetra_bd_color, stick_shininess);
						} else {
							if (cur_tetra->chiral == HNV_RIGHT_CHIRAL) {
								draw_tetra_tile_right(cur_tetra, tetra_open_color, tetra_open_color, open_shininess);
							} else {
								draw_tetra_tile_left(cur_tetra, tetra_open_color, tetra_open_color, open_shininess);
							}
							draw_cylinder_stick(cur_tetra, HNV_RADIUS, stick_open_color, stick_open_color, stick_shininess);
						}
					}
					
				} else if (visible_traj_type == HNV_UNKNOWN_T) {	// Show none
					if (cur_seq->type == HNV_CLOSED) {
						draw_cylinder_stick(cur_tetra, HNV_RADIUS_S, stick_closed_color, stick_closed_color, stick_shininess);
					} else {
						if (cur_tetra->belongs_to == HNV_BD) {
							draw_cylinder_stick(cur_tetra, HNV_RADIUS_S, tetra_bd_color, tetra_bd_color, bd_shininess);
						} else {
							draw_cylinder_stick(cur_tetra, HNV_RADIUS_S, stick_open_color, stick_open_color, stick_shininess);
						}
					}
					
				} else {						// Show all
					if (cur_seq->type == HNV_CLOSED) {
						if (cur_tetra->belongs_to == HNV_BD) {	// 2007-10-14
							if (cur_tetra->chiral == HNV_RIGHT_CHIRAL) {
								draw_tetra_tile_right(cur_tetra, tetra_bd_color, tetra_bd_color, bd_shininess);
							} else {
								draw_tetra_tile_left(cur_tetra, tetra_bd_color, tetra_bd_color, bd_shininess);
							}
							draw_cylinder_stick(cur_tetra, HNV_RADIUS, tetra_bd_color, tetra_bd_color, stick_shininess);
						} else {
							if (cur_tetra->chiral == HNV_RIGHT_CHIRAL) {
								draw_tetra_tile_right(cur_tetra, tetra_closed_color, tetra_closed_color, closed_shininess);
							} else {
								draw_tetra_tile_left(cur_tetra, tetra_closed_color, tetra_closed_color, closed_shininess);
							}
							draw_cylinder_stick(cur_tetra, HNV_RADIUS, stick_closed_color, stick_closed_color, stick_shininess);
						}
					} else {
						if (cur_tetra->belongs_to == HNV_BD) {
							if (cur_tetra->chiral == HNV_RIGHT_CHIRAL) {
								draw_tetra_tile_right(cur_tetra, tetra_bd_color, tetra_bd_color, bd_shininess);
							} else {
								draw_tetra_tile_left(cur_tetra, tetra_bd_color, tetra_bd_color, bd_shininess);
							}
							draw_cylinder_stick(cur_tetra, HNV_RADIUS, tetra_bd_color, tetra_bd_color, stick_shininess);
						} else {
							if (cur_tetra->chiral == HNV_RIGHT_CHIRAL) {
								draw_tetra_tile_right(cur_tetra, tetra_open_color, tetra_open_color, open_shininess);
							} else {
								draw_tetra_tile_left(cur_tetra, tetra_open_color, tetra_open_color, open_shininess);
							}
							draw_cylinder_stick(cur_tetra, HNV_RADIUS, stick_open_color, stick_open_color, stick_shininess);
						}
					}
					
				}
			}
		}
	} else if (visible_index < t_flow->seq_ct) {	// Show the specified trajectory only
		cur_seq = t_flow->sequences + visible_index;
		for (t_ct = 0; t_ct < cur_seq->length; t_ct++) {
			cur_tetra = cur_seq->tiles + t_ct;
			
			if (cur_seq->type == HNV_CLOSED) {
				if (cur_tetra->belongs_to == HNV_BD) {	// 2007-10-14
					if (cur_tetra->chiral == HNV_RIGHT_CHIRAL) {
						draw_tetra_tile_right(cur_tetra, tetra_bd_color, tetra_bd_color, bd_shininess);
					} else {
						draw_tetra_tile_left(cur_tetra, tetra_bd_color, tetra_bd_color, bd_shininess);
					}
					draw_cylinder_stick(cur_tetra, HNV_RADIUS, tetra_bd_color, tetra_bd_color, stick_shininess);
				} else {
					if (cur_tetra->chiral == HNV_RIGHT_CHIRAL) {
						draw_tetra_tile_right(cur_tetra, tetra_closed_color, tetra_closed_color, closed_shininess);
					} else {
						draw_tetra_tile_left(cur_tetra, tetra_closed_color, tetra_closed_color, closed_shininess);
					}
					draw_cylinder_stick(cur_tetra, HNV_RADIUS, stick_closed_color, stick_closed_color, stick_shininess);
				}
			} else {
				if (cur_tetra->belongs_to == HNV_BD) {
					if (cur_tetra->chiral == HNV_RIGHT_CHIRAL) {
						draw_tetra_tile_right(cur_tetra, tetra_bd_color, tetra_bd_color, bd_shininess);
					} else {
						draw_tetra_tile_left(cur_tetra, tetra_bd_color, tetra_bd_color, bd_shininess);
					}
					draw_cylinder_stick(cur_tetra, HNV_RADIUS_S, tetra_bd_color, tetra_bd_color, stick_shininess);
				} else {
					if (cur_tetra->chiral == HNV_RIGHT_CHIRAL) {
						draw_tetra_tile_right(cur_tetra, tetra_open_color, tetra_open_color, open_shininess);
					} else {
						draw_tetra_tile_left(cur_tetra, tetra_open_color, tetra_open_color, open_shininess);
					}
					draw_cylinder_stick(cur_tetra, HNV_RADIUS_S, stick_open_color, stick_open_color, stick_shininess);
				}
			}
		}
		for (seq_ct = 0; seq_ct < t_flow->seq_ct; seq_ct++) {
			cur_seq = t_flow->sequences + seq_ct;
			
			for (t_ct = 0; t_ct < cur_seq->length; t_ct++) {
				cur_tetra = cur_seq->tiles + t_ct;
				
				if (cur_seq->type == HNV_CLOSED) {
					draw_cylinder_stick(cur_tetra, HNV_RADIUS, stick_closed_color, stick_closed_color, stick_shininess);
				} else {
					if (cur_tetra->belongs_to == HNV_BD) {
						draw_cylinder_stick(cur_tetra, HNV_RADIUS_S, tetra_bd_color, tetra_bd_color, stick_shininess);
					} else {
						draw_cylinder_stick(cur_tetra, HNV_RADIUS_S, stick_open_color, stick_open_color, stick_shininess);
					}
				}
			}
		}
	} else {	// Show the peaks only
		for (seq_ct = 0; seq_ct < t_flow->seq_ct; seq_ct++) {
			cur_seq = t_flow->sequences + seq_ct;
			
			for (t_ct = 0; t_ct < cur_seq->length; t_ct++) {
				cur_tetra = cur_seq->tiles + t_ct;
				
				if (cur_seq->type == HNV_CLOSED) {
					draw_cylinder_stick(cur_tetra, HNV_RADIUS, stick_closed_color, stick_closed_color, stick_shininess);
				} else {
					if (cur_tetra->belongs_to == HNV_BD) {
						draw_cylinder_stick(cur_tetra, HNV_RADIUS_S, tetra_bd_color, tetra_bd_color, stick_shininess);
					} else {
						draw_cylinder_stick(cur_tetra, HNV_RADIUS_S, stick_open_color, stick_open_color, stick_shininess);
					}
				}
			}
		}
	}
	
	switch (visible_peak) {
		case HNV_STD_CONE:
			if (peak_visibility == ON) {
				for (peak_ct = 0; peak_ct < t_flow->std_cone_peak_size; peak_ct++) {
					if (t_flow->std_cone[peak_ct].redundance == HNV_NOT_REDUNDANT)
						draw_peak_sphere(t_flow->std_cone + peak_ct, blue, sphere_shininess, HNV_RADIUS_L);
				}
			} else {
				draw_peak_sphere(t_flow->std_cone + visible_peak_index, blue, sphere_shininess, HNV_RADIUS_L);
			}
			break;
		case HNV_CONJ_ROOF:
			for (peak_ct = 0; peak_ct < t_flow->conj_roof_peak_size; peak_ct++) {
				if (t_flow->conj_roof[peak_ct].redundance == HNV_NOT_REDUNDANT)
					draw_peak_sphere(t_flow->conj_roof + peak_ct, red, sphere_shininess, HNV_RADIUS_L);
			}
			break;
		case HNV_STD_ROOF:
			for (peak_ct = 0; peak_ct < t_flow->std_roof_peak_size; peak_ct++) {
				if (t_flow->std_roof[peak_ct].redundance == HNV_NOT_REDUNDANT)
					draw_peak_sphere(t_flow->std_roof + peak_ct, yellow, sphere_shininess, HNV_RADIUS_L);
			}
			break;
		default:
			break;
	}
	return;
}

